import path from 'path';
import fs from 'fs';
import dotenv from 'dotenv';
import * as commander from 'commander';
import packageJson from '../../package.json';
import {
  BUILD_DIR, DEVELOPMENT, DIST_DIR, PRODUCTION, STATIC_DIR, WEBSITE,
} from '../constants';
import { Scaffold } from './scaffold';
import { Core } from '../core';
import {
  getEnvAnalyzeServer,
  getEnvAnalyzeWeb,
  getEnvCache, getEnvEnvironment, getEnvHost, getEnvMode, getEnvPort, getEnvVars, getEnvVerbose,
} from '../util/env';

/**
 * Initial environment variables
 */
process.env.PAW_CACHE = '1';
process.env.PAW_VERBOSE = '0';
process.env.PAW_ANALYZE_WEB = '0';
process.env.PAW_ANALYZE_SERVER = '0';
process.env.PAW_ENV = DEVELOPMENT;
process.env.PAW_COMPILE_MODE = DEVELOPMENT;
process.env.PAW_PROJECT_VARS = JSON.stringify({});

export class Cli {
  // The commander program
  #program: commander.Command;

  // Project root this is basically the root from where the command is executed
  readonly #projectRoot: string;

  // the output directory of the server compiled files
  readonly #serverOutputPath: string;

  // the output directory of the server compiled files
  readonly #webOutputPath: string;

  // the base path of the web assets output like,
  // 'static'
  readonly #webAssetsBase: string;

  // The base path of server assets output, as we want it to be
  // in sync with web we put it in 'build' + the webAssetsBase
  readonly #serverAssetsBase: string;

  // The program name ReactPWA or PawJS
  readonly #programName: string;

  // The program bin name is how we can call it from cli,
  // ideally we keep it lowercase of programName after replacing space with
  // hyphen
  #programBinName: string;

  /**
   * Initialize the application with appropriate program name,
   * options and description
   */
  constructor() {
    this.#projectRoot = process.cwd();
    this.#serverOutputPath = path.join(this.#projectRoot, DIST_DIR);
    this.#webOutputPath = path.join(this.#projectRoot, DIST_DIR, BUILD_DIR);
    this.#webAssetsBase = STATIC_DIR;
    this.#serverAssetsBase = path.join(BUILD_DIR, this.#webAssetsBase);
    const executionScript = process?.argv?.[1] ?? '';
    const isExecutingViaReactPWA = executionScript.indexOf('reactpwa') !== -1;
    this.#programName = isExecutingViaReactPWA ? 'ReactPWA' : 'PawJS';
    this.#programBinName = this.#programName.toLowerCase();
    this.#program = new commander.Command(this.#programName);

    // Setup the name and description of the project
    this.setupProgram();

    // Add option for ENV
    this.optionEnv();

    // Add option for verbose execution
    this.optionVerbose();

    // Add option for non cached execution
    this.optionCache();

    // Add option for analyzing of webpack web modules
    this.optionAnalyze();

    // Add option to set mode of compilation
    this.optionMode();

    /**
     * Add commands for the program
     */
    // The start command
    this.commandStart().catch((ex) => {
      // eslint-disable-next-line no-console
      console.error('An error occurred with the start command', ex);
    });

    this.commandBuild().catch((ex) => {
      // eslint-disable-next-line no-console
      console.error('An error occurred with the start command', ex);
    });

    // the create command
    this.commandCreate().catch((ex) => {
      // eslint-disable-next-line no-console
      console.error('An error occurred with the create command', ex);
    });
  }

  /**
   * Set environment variables based on the program options.
   * this function should be called in all the commands so we have
   * latest values of the program options
   */
  setEnvVars() {
    /**
     * Read ENV file based on the PAW_ENV set which is affected by
     * --env variable from CLI options
     */
    const env = this.#program.opts().env?.toLowerCase() ?? DEVELOPMENT;
    const envFile = `.${env}.env`;

    const userEnvFilePath = path.join(this.#projectRoot, envFile);
    let result;
    if (fs.existsSync(userEnvFilePath)) {
      result = dotenv.config({
        path: userEnvFilePath,
      });
    } else {
      result = dotenv.config();
    }
    process.env.PAW_PROJECT_VARS = JSON.stringify(result?.parsed ?? {});
  }

  /**
   * Setup the name of program and the bin name
   */
  setupProgram() {
    this.#program.name(this.#programBinName)
      .version(packageJson.version)
      .description(`${this.#programName} is a framework that can help you create Progressive Web Applications with ReactJS.`
        + ` It has been built with SEO and performance in mind. Visit ${WEBSITE} for detailed documentation`);
  }

  /**
   * Set global option env for the program
   */
  optionEnv() {
    this.#program.option(
      '-e, --env <env>',
      `Set the application environment. defaults to '${DEVELOPMENT}' environment`,
      DEVELOPMENT,
    ).on('option:env', () => {
      let env = DEVELOPMENT;
      if (this.#program.opts().env && typeof this.#program.opts().env === 'string') {
        env = this.#program.opts().env.toLowerCase();
        if (env === 'dev') {
          // eslint-disable-next-line no-console
          console.info(`NOTE:: Setting env to ${DEVELOPMENT}. Please use --env=${DEVELOPMENT} instead`);
          env = DEVELOPMENT;
        }
        if (env === 'prod') {
          // eslint-disable-next-line no-console
          console.info(`NOTE:: Setting env to ${PRODUCTION}. Please use --env=${PRODUCTION} instead`);
          env = PRODUCTION;
        }
      }
      if (env === PRODUCTION) {
        process.env.PAW_COMPILE_MODE = env;
      }
      process.env.PAW_ENV = env;
    });
  }

  optionVerbose() {
    /**
     * Set global option verbose for the program
     */
    this.#program.option(
      '-v, --verbose',
      'Start with detailed comments and explanation',
      false,
    ).on('option:verbose', () => {
      process.env.PAW_VERBOSE = this.#program.opts().verbose ? '1' : '0';
    });
  }

  optionAnalyze() {
    /**
     * Set global option analyze web and analyze server for the program
     */
    this.#program.option(
      '--analyze-web',
      'Analyze your webpack components and code splitting for client generated code',
      false,
    ).on('option:analyze-web', () => {
      process.env.PAW_ANALYZE_WEB = this.#program.opts().analyzeWeb ? '1' : '0';
    });

    this.#program.option(
      '--analyze-server',
      'Analyze your webpack components for server generated code',
      false,
    ).on('option:analyze-server', () => {
      process.env.PAW_ANALYZE_SERVER = this.#program.opts().analyzeServer ? '1' : '0';
    });
  }

  optionMode() {
    this.#program.option(
      '-m, --mode <mode>',
      `Set the application compile mode, which is in turn used by webpack. defaults to '${DEVELOPMENT}' but it can be either '${DEVELOPMENT}' or '${PRODUCTION}'`,
      DEVELOPMENT,
    ).on('option:mode', () => {
      let mode = DEVELOPMENT;
      if (this.#program.opts().mode && typeof this.#program.opts().mode === 'string') {
        mode = this.#program.opts().env.toLowerCase();
        if (mode === 'dev') {
          // eslint-disable-next-line no-console
          console.info(`NOTE:: Setting compilation mode to ${DEVELOPMENT}. Please use --mode=${DEVELOPMENT} instead`);
          mode = DEVELOPMENT;
        } else if (mode === 'prod') {
          // eslint-disable-next-line no-console
          console.info(`NOTE:: Setting compilation mode to ${PRODUCTION}. Please use --env=${PRODUCTION} instead`);
          mode = PRODUCTION;
        } else if (mode !== DEVELOPMENT && mode !== PRODUCTION) {
          // eslint-disable-next-line no-console
          console.error(`ERROR:: Invalid compilation mode '${mode}' provided. It can be either '${PRODUCTION}' or '${DEVELOPMENT}' only, thus changing it to ${DEVELOPMENT}`);
          mode = DEVELOPMENT;
        }
      }
      process.env.PAW_COMPILE_MODE = mode;
    });
  }

  /**
   * Add option to disable cache while development, this will disable all caching mechanism with
   * current application along with babel cache
   */
  optionCache() {
    this.#program
      .option(
        '--no-cache',
        `Disable cache. Ideally used for ${this.#programName} core/plugin ${DEVELOPMENT}. If you are using `
        + `${this.#programName} for your own project development we do not encourage you to disable cache.`,
        false,
      )
      .on('option:no-cache', () => {
        const doCache = this.#program.opts()?.cache ? '1' : '0';
        process.env.BABEL_DISABLE_CACHE = doCache;
        process.env.PAW_CACHE = doCache;
      });
  }

  /**
   * Start the development of the repository
   */
  async commandStart() {
    this.#program.command('start')
      .description(`Start the program in ${DEVELOPMENT} mode`)
      .action(async () => {
        this.setEnvVars();
        const core = new Core({
          cache: getEnvCache(),
          verbose: getEnvVerbose(),
          env: getEnvEnvironment(),
          mode: getEnvMode(),
          envVars: getEnvVars(),
        });
        /**
         * Create compiler instance for the server
         */
        const serverCompiler = core.compileServer({
          outputPath: this.#serverOutputPath,
          assetsBase: this.#serverAssetsBase,
          analyze: getEnvAnalyzeServer(),
        });
        const webCompiler = core.compileWeb({
          outputPath: this.#webOutputPath,
          assetsBase: this.#webAssetsBase,
          analyze: getEnvAnalyzeWeb(),
        });
        await core.startServer(
          serverCompiler,
          webCompiler,
          {
            port: getEnvPort(),
            host: getEnvHost(),
          },
        );
      });
  }

  /**
   * Start the development of the repository
   */
  async commandBuild() {
    this.#program.command('build')
      .description('Build you current code codebase')
      .action(async () => {
        this.setEnvVars();
        const core = new Core({
          cache: getEnvCache(),
          verbose: getEnvVerbose(),
          env: getEnvEnvironment(),
          mode: getEnvMode(),
          envVars: getEnvVars(),
        });
        /**
         * Create compiler instance for the server
         */
        const serverCompiler = core.compileServer({
          outputPath: this.#serverOutputPath,
          assetsBase: this.#serverAssetsBase,
          analyze: getEnvAnalyzeServer(),
        });
        const webCompiler = core.compileWeb({
          outputPath: this.#webOutputPath,
          assetsBase: this.#webAssetsBase,
          extractCSS: true,
          analyze: getEnvAnalyzeWeb(),
        });
        await core.run(serverCompiler, { watch: false });
        await core.run(webCompiler, { watch: false });
      });
  }

  /**
   * Create a boilerplate from scratch with the create function
   */
  async commandCreate() {
    this.#program
      .command('create <name>')
      .option(
        '--no-typescript',
        'Scaffold project with default as JavaScript instead of TypeScript',
        false,
      )
      .description(`Scaffold a new project with ${this.#programName}.`)
      .action(async (name, options) => {
        this.setEnvVars();
        const scaffoldingProject = new Scaffold({
          projectName: name,
          withTypeScript: options?.typescript ?? false,
        });
        try {
          scaffoldingProject.createProjectFolder();
          await scaffoldingProject.copyProjectFiles();
          await scaffoldingProject.installPackages();
        } catch (ex) {
          // eslint-disable-next-line no-console
          console.error('Error occurred while creating the project', ex);
        }
      });
  }

  init() {
    this.#program.parse(process.argv);
    // If no commands are specified display help
    if (process.argv.length < 3) {
      this.#program.help();
    }
  }
}
