import path from 'path';
import fs from 'fs';
import { execSync } from 'child_process';
import { copyFiles } from '../util/common';
import { getReplacementVars } from '../util/replacement-vars';

const currentDir = path.resolve(process.cwd());

type ScaffoldingOptions = {
  projectName: string;
  withTypeScript: boolean;
};

class Scaffold {
  readonly #projectPath: string;

  readonly #projectName: string;

  readonly #withTypeScript: boolean;

  constructor(
    {
      projectName: projectNameWithPath,
      withTypeScript,
    }: ScaffoldingOptions,
  ) {
    let projectPath: string = './';
    let projectName: string = projectNameWithPath;
    if (projectName.indexOf(path.sep) !== -1) {
      const splitPath = projectNameWithPath.split(path.sep);
      projectName = `${splitPath.pop()}`;
      projectPath = `${splitPath.join(path.sep)}`;
    }
    this.#projectName = projectName;
    this.#projectPath = projectPath;
    this.#withTypeScript = withTypeScript;
  }

  createProjectFolder(): string {
    let absolutePath = path.join(this.#projectPath, this.#projectName);
    if (!path.isAbsolute(absolutePath)) {
      absolutePath = path.resolve(path.join(currentDir, absolutePath));
    }
    // Check if the parent dir is writable
    const baseDirectory = path.dirname(absolutePath);
    try {
      // eslint-disable-next-line no-bitwise
      fs.accessSync(baseDirectory, fs.constants.R_OK | fs.constants.W_OK);
    } catch (err) {
      // eslint-disable-next-line no-console
      console.error(`ERR: Cannot create the project at ${absolutePath}. Check if you have appropriate permission for doing so.`);
      throw Error('Cannot create project directory');
    }

    let folderAlreadyExists = false;
    try {
      // eslint-disable-next-line no-bitwise
      fs.accessSync(absolutePath, fs.constants.R_OK | fs.constants.W_OK);
      // eslint-disable-next-line no-console
      console.error(`ERR: Cannot proceed as a folder already exists at ${absolutePath}.`);
      folderAlreadyExists = true;
    } catch (ex) {
      // do nothing here as this should be expected
    }
    if (folderAlreadyExists) {
      throw Error(`Cannot proceed as the path already exists: ${absolutePath}`);
    }

    try {
      fs.mkdirSync(absolutePath);
    } catch (ex) {
      // eslint-disable-next-line no-console
      console.error('ERR: An error occurred while creating the project folder', ex);
      throw Error(`Error occurred while creating the project directory: ${absolutePath}`);
    }
    return absolutePath;
  }

  async copyProjectFiles() {
    const destination = path.resolve(path.join(this.#projectPath, this.#projectName));
    const rootSource = path.resolve(
      path.join(
        __dirname,
        '..',
        'scaffold',
      ),
    );
    const srcSource = path.resolve(
      path.join(
        rootSource,
        this.#withTypeScript ? 'typescript' : 'javascript',
      ),
    );
    const replacementVars = getReplacementVars(this.#projectName, this.#projectPath);

    await copyFiles(srcSource, destination, {
      replacementVars,
      withSubDirectory: true,
    });
    await copyFiles(rootSource, destination, {
      replacementVars,
      withSubDirectory: false,
    });
  }

  async installPackages() {
    const absolutePath = path.join(this.#projectPath, this.#projectName);
    try {
      execSync(
        'npm install',
        {
          cwd: absolutePath,
          stdio: 'inherit',
        },
      );
      return true;
    } catch (ex) {
      // eslint-disable-next-line no-console
      console.error('Error occurred while installing packages', ex);
    }
    return false;
  }
}

export { Scaffold };
