import 'regenerator-runtime/runtime';

/**
 * Steps to do while importing the initial entry.tsx
 * - On document load, import client handler
 * - Import client from project root
 * - Import router if any and do preloading
 */
let loadApplicationTimeout = 0;

const loadApplication = async (deadline: RequestIdleCallbackParams) => {
  let deadlineFallback = deadline;
  // If there is no deadline, just run as long as necessary.
  if (typeof deadline === 'undefined') {
    deadlineFallback = {
      didTimeout: false,
      timeRemaining() { return Number.MAX_VALUE; },
    };
  }
  if ((deadlineFallback.timeRemaining() > 0 || deadlineFallback.didTimeout)) {
    // Do something over here
    /**
     * - Import client handler async
     * - Import project client async
     * - Pass the project client to client handler
     * - Load the application
     */
    const { ClientHandler } = await (() => import('./handler'))();
    const handlerInstance = new ClientHandler();
    handlerInstance.run();
  } else {
    if (loadApplicationTimeout) {
      window.cancelIdleCallback(loadApplicationTimeout);
    }
    loadApplicationTimeout = window.requestIdleCallback(loadApplication, { timeout: 2000 });
  }
};

window.requestIdleCallback = window.requestIdleCallback
  || function requestIdleCallback(cb, options) {
    return setTimeout(() => {
      const start = Date.now();
      cb({
        didTimeout: false,
        timeRemaining() {
          return Math.max(0, 50 - (Date.now() - start));
        },
      });
    }, typeof options !== 'undefined' ? options?.timeout ?? 0 : 0);
  };

window.cancelIdleCallback = window.cancelIdleCallback
  || function cancelIdleCallback(id: any) {
    clearTimeout(id);
  };

// If script is loaded via async method, the readyState of document is either
// complete or interactive and not loading, thus the event DOMContentLoaded does
// not fire.
if (document.readyState !== 'loading') {
  loadApplicationTimeout = window.requestIdleCallback(loadApplication, { timeout: 2000 });
} else {
  document.addEventListener('DOMContentLoaded', () => {
    loadApplicationTimeout = window.requestIdleCallback(loadApplication, { timeout: 2000 });
  });
}
