import React, {
  Context,
  createContext, FC, PropsWithChildren, useContext, useRef,
} from 'react';
import serialize from 'serialize-javascript';
import { suspensify } from './suspensify';

const DataContext: Context<any> = createContext(null);
const PreloaderContext: Context<any> = createContext(null);

export const PreloaderProvider: FC<{ context?: any }> = (
  { children, context }: PropsWithChildren<{ context?: any }>,
) => {
  const preloadRef: any = useRef(context ?? {});
  const addPreloadData = (cb: any, id: string) => {
    preloadRef.current[id] = cb;
  };
  const getPreloadData = (id: string) => preloadRef.current?.[id] ?? null;
  return (
    <PreloaderContext.Provider value={{ addPreloadData, getPreloadData, ref: preloadRef.current }}>
      {children}
    </PreloaderContext.Provider>
  );
};

export const DataProvider: FC = (
  { children }: PropsWithChildren<{}>,
) => {
  const { addPreloadData, getPreloadData } = useContext(PreloaderContext);
  const context: any = useRef({});

  const setSuspenseCallback = (cb: any, id = '', reUse = false) => {
    if (context?.current?.[id]) {
      context.current[id] = null;
    }
    if (id && reUse) {
      const data = getPreloadData(id);
      if (data && context?.current?.[id]) {
        throw new Error(`Duplicate Preload ID: ${id}!`);
      } else if (data) {
        context.current[id] = {
          read: () => data,
        };
        return;
      }
    }
    const dynamicCb = suspensify((typeof cb === 'function' ? cb() : cb) ?? {});
    if (id && reUse) {
      addPreloadData(dynamicCb, id);
    }
    context.current[id] = dynamicCb;
  };

  const getSuspenseCallback = (id = '', reUse = false) => {
    const data = reUse ? getPreloadData(id) : null;
    return data ?? context.current?.[id] ?? null;
  };

  return (
    <DataContext.Provider value={{ setSuspenseCallback, getSuspenseCallback }}>
      {children}
    </DataContext.Provider>
  );
};

type UseDataOptions = {
  id?: string;
  reUse?: boolean;
};
export const useData = (arg: any, options: UseDataOptions = {}) => {
  const { setSuspenseCallback, getSuspenseCallback } = useContext(DataContext);
  const reUse = !!(options?.reUse);
  if (reUse && typeof arg !== 'function') {
    throw new Error('If you are planning to reuse the data from server, then please pass the argument as a function, example useData(() => Promise<any>, { reUse: true, id: <id> })');
  }
  if (reUse && !options?.id) {
    throw new Error('If you are planning to reuse the data from server, then please set the unique id for the data as that will be used as reference when hydrating. example useData(() => Promise<any>, { reUse: true, id: <id> })');
  }
  if (!getSuspenseCallback(options?.id, reUse)) {
    setSuspenseCallback(arg, options?.id, reUse);
  }
  return getSuspenseCallback(options?.id, reUse)?.read();
};

export const PreloaderData: FC = ({ children }: PropsWithChildren<{}>) => {
  const { ref } = useContext(PreloaderContext);
  const data = useData(Promise.resolve(ref));
  const keys = Object.keys(data);
  const preloadedData: any = {};
  if (keys.length) {
    keys.forEach((key) => {
      preloadedData[key] = data[key].read();
    });
  }
  return (
    <>
      {children}
      <script
        key="PreloadDataScript"
        // eslint-disable-next-line react/no-danger
        dangerouslySetInnerHTML={{
          __html: `window.pawPreloadedData = ${serialize(preloadedData)}`,
        }}
      />
    </>
  );
};
