import webpack from 'webpack';
import path from 'path';
import { webConfig } from '../webpack/web.config';
import { serverConfig } from '../webpack/server.config';
import { DEVELOPMENT } from '../constants';
import { DevelopmentServer } from '../server/development';
import {
  ConfigOptions,
  CoreOptions,
  RunOptions,
  ServerOptions,
} from '../@types/core';

class Core {
  /**
   * @description Option to set if we want to cache the bundling
   * process. Ideally the cache value should be true, it should not affect
   * the application, until and unless developing or working with webpack plugins
   * @private #cache boolean
   */
  readonly #cache: boolean;

  /**
   * @description Option to get more compilation details
   * @private #verbose boolean
   */
  readonly #verbose: boolean;

  /**
   * @description Set the environment of the application as desired,
   * this can be used for various reason
   * like differentiating local, staging, development & production. Ideal values are
   * "DEVELOPMENT" or "PRODUCTION"
   * @private #env string
   */
  readonly #env: string;

  /**
   * @description The compilation mode of the application. This mode is passed on
   * to the webpack compiler with option 'mode'.
   * Valid enum values are "production", "development" and "none"
   * @private #mode enum('development', 'production', 'none')
   */
  readonly #mode: 'development' | 'production' | 'none';

  /**
   * @description When targeting to host the application on a
   * Content delivery network, specify the URL Prefix from where the
   * application will be served.
   * @example If the CDN url is https://res.cloudinary.com/atyantik/raw/upload/js/8f9b299e8d83241aa22f.js
   * then the CDN URL is a `https://res.cloudinary.com/atyantik/raw/upload/`
   * @private #cdnUrl string
   */
  readonly #cdnUrl: string;

  /**
   * @description App url is very much required when you are loading the application
   * via a sub directory or have a prefix in the url
   * so loading your app side by side your wordpress application on url
   * https://www.reactpwa.com/my-app/, the application url is '/my-app/'
   * @private #appUrl string
   */
  readonly #appUrl: string;

  /**
   * The root/context of the project. Usually it is the root of the current project
   * if nothing is provided the default is process.cwd()
   * @private #root string
   */
  readonly #root: string;

  readonly #projectVars: { [key: string]: any };

  constructor(options: CoreOptions) {
    this.#cache = options?.cache ?? true;
    this.#verbose = options?.verbose ?? false;
    this.#env = options?.env ?? DEVELOPMENT;
    this.#mode = options?.mode ?? DEVELOPMENT;
    this.#cdnUrl = options?.cdnUrl ?? '';
    this.#appUrl = options?.appUrl ?? '/';
    this.#root = options?.root ?? process.cwd();
    this.#projectVars = Object.freeze(options?.envVars ?? {});

    /**
     * Set the project root with absolute path
     */
    if (!path.isAbsolute(this.#root)) {
      this.#root = path.resolve(process.cwd(), this.#root);
      if (!this.#root) {
        // eslint-disable-next-line no-console
        console.error(`Cannot resolve root: ${options?.root ?? this.#root}, Thus setting it to ${process.cwd()}`);
        this.#root = process.cwd();
      }
    }

    /**
     * Update the appUrl with a leading '/'
     */
    if (!this.#appUrl.endsWith('/')) {
      this.#appUrl += '/';
    }
  }

  private getDefaultCompileOptions: () => ConfigOptions = () => ({
    mode: this.#mode,
    env: this.#env,
    verbose: this.#verbose,
    cache: this.#cache,
    resourceUrl: this.getResourceUrl(),
    root: this.#root,
    envVars: this.#projectVars,
  });

  // eslint-disable-next-line class-methods-use-this
  compile(config: webpack.Configuration) {
    return webpack(config, Core.compileCallbackHandler(config));
  }

  getServerConfig(options: ConfigOptions) {
    return serverConfig({
      ...this.getDefaultCompileOptions(),
      ...options,
    });
  }

  compileServer(options: ConfigOptions) {
    const config = this.getServerConfig(options);
    return webpack(config);
  }

  getWebConfig(options: ConfigOptions) {
    return webConfig({
      ...this.getDefaultCompileOptions(),
      ...options,
    });
  }

  compileWeb(options: ConfigOptions) {
    const config = this.getWebConfig(options);
    return webpack(config);
  }

  /**
   * @description Resource Url is the base url from where the content is served.
   * So if the content is served from a CDN, the resource URL is the CDN URL,
   * but if there is no CDN URL and the app is served from a sub directory,
   * we use the APP_URL to serve the content from
   */
  public getResourceUrl() {
    let resourceUrl = this.#cdnUrl;
    if (!resourceUrl || resourceUrl === '' || resourceUrl === '/') {
      resourceUrl = this.#appUrl;
    }
    return resourceUrl;
  }

  /**
   * @description Compile a callback handler for the configuration of web or server
   * @param options webpack.WebpackOptionsNormalized
   */
  static compileCallbackHandler = (
    options: webpack.WebpackOptionsNormalized | webpack.Configuration,
  ) => (err: Error | undefined, stats: webpack.Stats | webpack.MultiStats | undefined) => {
    if (err || stats?.hasErrors()) {
      // Handle error here if necessary
    }
    if (err) {
      // eslint-disable-next-line no-console
      console.error(err);
    }
    if (stats) {
      // eslint-disable-next-line no-console
      console.info(stats.toString(options.stats));
    }
  };

  /**
   * @description Run web compiler with the RunOptions,
   * it can be a single time compilation or continuous compilation.
   * apparently when creating this library webpack-dev-middleware and webpack-hot-middleware
   * have issues with webpack.MultiCompiler, thus I am trying to avoid it completely.
   * @param compiler webpack.Compiler
   * @param options RunOptions
   * @return this Core
   */
  async run(compiler: webpack.Compiler, options: RunOptions) {
    let resolved = false;
    return new Promise((resolve) => {
      if (options?.watch) {
        compiler.watch({
          aggregateTimeout: 200,
          poll: 1000,
          ignored: /node_modules/,
        }, Core.compileCallbackHandler(compiler.options));
      } else {
        compiler.run(Core.compileCallbackHandler(compiler.options));
      }
      compiler.hooks.afterDone.tap('ResolveCompilationDone', () => {
        if (!resolved) {
          resolved = true;
          resolve(this);
        }
      });
    });
  }

  async startServer(
    serverCompiler: webpack.Compiler,
    webCompiler: webpack.Compiler,
    serverOptions: ServerOptions,
  ) {
    const server = new DevelopmentServer({ serverCompiler, webCompiler });
    await server.start(serverOptions);
    return this;
  }
}
export { Core };
