import path from 'path';
import { Compiler } from 'webpack';
import express, { Request, Response, NextFunction } from 'express';
import webpackMiddleware from 'webpack-dev-middleware';
import ora from 'ora';
import { ServerCompilers, ServerOptions } from '../@types/core';
import { requireFromString } from '../util/common';
import { normalizeAssets } from '../common/normalize';

class DevelopmentServer {
  readonly #webCompiler: Compiler;

  readonly #serverCompiler: Compiler;

  readonly #app: express.Application;

  #serverCompiled: boolean = false;

  #webCompiled: boolean = false;

  #serverStarted: boolean = false;

  #spinner;

  constructor({ webCompiler, serverCompiler }: ServerCompilers) {
    this.#webCompiler = webCompiler;
    this.#serverCompiler = serverCompiler;

    // Initialize express application
    this.#app = express();
    this.#spinner = ora();
  }

  // eslint-disable-next-line class-methods-use-this
  #handleRequests(
    serverMiddleware: webpackMiddleware.WebpackDevMiddleware,
    webMiddleware: webpackMiddleware.WebpackDevMiddleware,
  ) {
    return (
      req: Request,
      res: Response,
      next: NextFunction,
    ) => {
      let publicPath = serverMiddleware?.context?.stats?.toJson?.()?.publicPath ?? '';
      if (publicPath.endsWith('/')) {
        publicPath = publicPath.substring(0, publicPath.length - 1);
      }

      // @ts-ignore
      const { outputFileSystem: serverOutputFileSystem } = serverMiddleware.context;
      // @ts-ignore
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      const { outputFileSystem: webOutputFileSystem } = webMiddleware.context;

      const assets = normalizeAssets(webMiddleware?.context?.stats?.toJson?.() ?? {});

      // Read Development File
      const serverFile = serverMiddleware.getFilenameFromUrl(
        `${publicPath}/server.js`,
      );
      if (!serverFile) {
        throw new Error(`Cannot find server.js at ${publicPath}/server.js.`
          + ' Must be some error with compilation use --verbose mode to trace the error');
      }

      const compiledServerContent = serverOutputFileSystem.readFileSync(serverFile, { encoding: 'utf-8' });

      const nodePath = process.env.NODE_PATH || '';
      try {
        const compiledServerModule = requireFromString(compiledServerContent, {
          appendPaths: nodePath.split(path.delimiter),
        });
        if (compiledServerModule?.server) {
          compiledServerModule.server.setAssets(assets);
          compiledServerModule.server.setPublicPath(publicPath);
          compiledServerModule.server.handleRequests();
          return compiledServerModule.server.getExpressApplication().handle(req, res, next);
        }
      } catch (ex) {
        // eslint-disable-next-line no-console
        console.error('An unexpected error occurred:', ex);
      }
      return next();
    };
  }

  #startOnCompile(
    {
      serverMiddleware,
      webMiddleware,
      host,
      port,
    }: ServerOptions & {
      serverMiddleware: webpackMiddleware.WebpackDevMiddleware,
      webMiddleware: webpackMiddleware.WebpackDevMiddleware,
    },
  ) {
    if (
      this.#serverCompiled
      && this.#webCompiled
      && !this.#serverStarted
    ) {
      this.#serverStarted = true;
      this.#app.use(this.#handleRequests(serverMiddleware, webMiddleware));
      this.#app.listen(port, host, () => {
        this.#spinner.stop();
        // eslint-disable-next-line no-console
        console.log(`\n> Ready on http://${host}:${port}`);
      });
    }
    return this.#app;
  }

  async start(options: ServerOptions) {
    // eslint-disable-next-line no-console
    console.clear();
    this.#spinner.start('Bundling modules...');
    /**
     * Add server middleware
     */
    const serverMiddleware = webpackMiddleware(this.#serverCompiler, {
      writeToDisk: false,
      serverSideRender: true,
    });
    // Add the serverWebpackMiddleware to the express application
    this.#app.use(serverMiddleware);

    /**
     * Add Web Middleware
     */
    const webMiddleware = webpackMiddleware(this.#webCompiler, {
      writeToDisk: false,
      serverSideRender: true,
    });
    this.#app.use(webMiddleware);

    /**
     * Check for server compilation
     */
    this.#serverCompiler.hooks.afterDone.tap(
      'StartServerOnServerCompilation',
      () => {
        this.#spinner.clear();
        this.#serverCompiled = true;
        this.#startOnCompile(
          {
            serverMiddleware,
            webMiddleware,
            ...options,
          },
        );
      },
    );

    /**
     * Check for web compilation
     */
    this.#webCompiler.hooks.afterDone.tap(
      'StartServerOnWebCompilation',
      () => {
        this.#spinner.clear();
        this.#webCompiled = true;
        this.#startOnCompile(
          {
            serverMiddleware,
            webMiddleware,
            ...options,
          },
        );
      },
    );
  }
}
export { DevelopmentServer };
