import React from 'react';
import { URL } from 'url';
import express from 'express';
import compression from 'compression';
import { ApplicationRequestHandler } from 'express-serve-static-core';
// @ts-ignore
import { pipeToNodeWritable } from 'react-dom/server';
import { StaticRouter as Router } from 'react-router-dom/server';
import ProjectRoutes from 'pawProjectRoutes';
import { RouteObject } from 'react-router';
import { App } from '../components/app';
import { Html } from '../components/html';
import { ServerOptions } from '../@types/core';
import { RoutesHandler } from '../router/handler';
import { PreloaderProvider } from '../components/data';
import { Route } from '../@types/router';

const addFullUrl = (
  req: express.Request,
  res: express.Response,
  next: express.NextFunction,
) => {
  try {
    const protocol = req.get('X-Forwarded-Protocol')
      || req.get('X-Forwarded-Proto')
      || req.protocol
      || (req.secure ? 'https' : 'http');
    const host = req.get('X-Host') || req.get('X-Forwarded-Host') || req.get('host');
    const parsedUrl = new URL(req.originalUrl, `${protocol}://${host}`);
    let pathName = '/';
    if (parsedUrl.pathname) {
      pathName = parsedUrl.pathname;
      if (!pathName.endsWith('/')) {
        pathName += '/';
      }
    }
    res.locals.fullUrl = `${protocol}://${host}${pathName}${parsedUrl.search || ''}`;
  } catch (ex) {
    // Some error with parsing of url
  }
  return next();
};

class StandaloneServer {
  #routes: Route[] = [];

  #compiledRoutes: RouteObject[] = [];

  #projectRoutes: ProjectRoutes;

  readonly #app: express.Application;

  #assets: string[];

  #publicPath: string;

  addMiddleware: ApplicationRequestHandler<express.Application>;

  #HtmlComponent: React.FC<any>;

  constructor() {
    this.#app = express();
    this.#app.disable('x-powered-by');
    /**
     * Enable trust proxy, i.e. allow data from
     * X-Protocol
     * X-Host
     * X-Forwarded-For
     */
    this.#app.enable('trust proxy');
    this.#assets = [];
    this.#publicPath = '';
    this.addMiddleware = this.#app.use.bind(this.#app);
    this.#HtmlComponent = Html;

    // Add compression by default
    this.addMiddleware(compression());
    this.addMiddleware(addFullUrl);
    this.#projectRoutes = new ProjectRoutes();
  }

  async initRoutes(req: express.Request, res: express.Response) {
    const url = new URL(res.locals.fullUrl);
    const userAgent = req.get('User-Agent') ?? '';
    const languages = req.acceptsLanguages() ?? [];
    const language = languages.length ? languages[0] : '';
    const returnValue: Route [] | Promise<Route []> = this
      .#projectRoutes
      ?.getRoutes?.({
        url,
        userAgent,
        languages,
        language,
      });

    this.#routes = await returnValue;
    const routeHandler = new RoutesHandler();
    this.#compiledRoutes = routeHandler.compileRoutes(this.#routes);
  }

  setAssets(assets: string[]) {
    this.#assets = [...assets];
  }

  setPublicPath(publicPath: string) {
    this.#publicPath = publicPath;
  }

  setHtmlComponent(component: React.FC<any>) {
    this.#HtmlComponent = component;
  }

  handleRequests() {
    this.#app.use(
      async (
        req: express.Request,
        res: express.Response,
      ) => {
        // The new wiring is a bit more involved.
        res?.socket?.on?.('error', (error) => {
          // eslint-disable-next-line no-console
          console.error('Fatal', error);
        });
        /**
         * Get request path with public path appended at start
         */
        let publicRequestPath = this.#publicPath;
        if (!req.path.startsWith('/')) {
          publicRequestPath += '/';
        }
        publicRequestPath += req.path;

        // @todo: handle this more properly. Allow user to view favicon in start mode
        // If the request is for static file favicon, do not compute or send data to
        // server just execute the express default next functionality
        // to let it manage itself.
        if (
          // if the request is for favicon
          publicRequestPath.indexOf('favicon.ico') !== -1
        ) {
          return res.status(404).send();
        }

        await this.initRoutes(req, res);

        let didError = false;
        const element = (
          <PreloaderProvider>
            <Router location={req.url}>
              <App
                assets={this.#assets}
                routes={this.#compiledRoutes}
              />
            </Router>
          </PreloaderProvider>
        );
        // const HtmlComponent: React.FC<any> = this.#HtmlComponent;
        const { startWriting } = pipeToNodeWritable(
          element,
          res,
          {
            onReadyToStream() {
              // If something errored before we started streaming,
              // we set the error code appropriately.
              res.statusCode = didError ? 500 : 200;
              res.setHeader('Content-type', 'text/html');
              res.write('<!DOCTYPE html>');
              startWriting();
            },
            onError(x: Error) {
              didError = true;
              // eslint-disable-next-line no-console
              console.error(x);
            },
          },
        );
        return null;
      },
    );
  }

  start({ port, host }: ServerOptions) {
    this.#app.listen(port, host, () => {
      // eslint-disable-next-line no-console
      console.log(`> Ready on http://${host}:${port}`);
    });
  }

  getExpressApplication() {
    return this.#app;
  }
}

export { StandaloneServer };
