import fs from 'fs';
import path from 'path';
import Module from 'module';

const falseStringValues = [
  'false',
  '0',
  'null',
  'undefined',
  '{}',
  '[]',
  'NaN',
  '',
];
const envReturnType = {
  String: 'string',
  Boolean: 'boolean',
  Number: 'number',
} as const;

type ValuesOf<T> = T[keyof T];
type ReturnType = ValuesOf<typeof envReturnType>;

export const getEnv = (
  envVar: string,
  defaultValue = '',
  type: ReturnType = envReturnType.String,
) => {
  let val = defaultValue;
  if (envVar && typeof process.env[envVar] !== 'undefined') {
    val = `${process.env[envVar]}`;
  }
  switch (type) {
    case envReturnType.Boolean:
      return falseStringValues.indexOf(val.toLowerCase()) === -1;
    case envReturnType.String:
    default:
      return val;
    case envReturnType.Number:
      return Number.isNaN(val) ? 0 : Number.parseFloat(val);
  }
};

export const copyFiles = async (
  source: string,
  destination: string,
  options: {
    withSubDirectory?: boolean,
    replacementVars?: {
      [key: string]: string,
    }
  },
) => {
  const replacementVars = options?.replacementVars ?? {};
  const replacementVarKeys = Object.keys(replacementVars);
  try {
    // Get the files as an array
    const files = await fs.promises.readdir(source);

    // Loop them all with the new for...of
    for (let fileIndex = 0; fileIndex < files.length; fileIndex += 1) {
      const file = files[fileIndex];
      const fromPath = path.join(source, file);
      const toPath = path.join(destination, file);
      // eslint-disable-next-line no-await-in-loop
      const stat = await fs.promises.stat(fromPath);
      if (stat.isFile()) {
        // eslint-disable-next-line no-await-in-loop
        let fileContent = await fs.promises.readFile(fromPath, { encoding: 'utf-8' });
        replacementVarKeys.forEach((key) => {
          fileContent = fileContent
            .replace(/\/\/ @ts-ignore\n/ig, '')
            .replace(new RegExp(`{${key}}`, 'ig'), replacementVars[key]);
        });
        fs.writeFileSync(toPath, fileContent, { encoding: 'utf-8' });
      } else if (stat.isDirectory() && options.withSubDirectory) {
        // eslint-disable-next-line no-await-in-loop
        await fs.promises.mkdir(toPath, {
          recursive: true,
        });
        // await fs.promises.rename(fromPath, toPath);
        // eslint-disable-next-line no-await-in-loop
        await copyFiles(fromPath, toPath, options);
      }
    }
  } catch (e) {
    // Catch anything bad that happens
    // eslint-disable-next-line no-console
    console.error('We\'ve thrown! Whoops!', e);
  }
};

const supportedExtensions = [
  'js',
  'jsx',
  'ts',
  'tsx',
];

/**
 * We need to resolve the files as per the extensions at many places
 * for example we do not want to restrict people to just .js or .jsx extension
 * we need ability like fileExistsSync to compare for all extensions we have defined
 * in `src/extensions`
 * @type {*|Function}
 */
export const pawExistsSync = ((filePath: string) => {
  if (fs.existsSync(filePath)) return filePath;
  let resolvedFilePath = '';
  supportedExtensions.forEach((jsExt: string) => {
    if (resolvedFilePath) {
      return;
    }
    const extensionPath = `${filePath}.${jsExt}`;
    if (fs.existsSync(extensionPath)) {
      resolvedFilePath = extensionPath;
    }
  });
  return resolvedFilePath;
});

export const requireFromString = (code: string, file: any, options?: any) => {
  let opts = options;
  let filename = file;
  if (typeof filename === 'object') {
    opts = filename;
    filename = undefined;
  }

  opts = opts || {};
  filename = filename || '';

  opts.appendPaths = opts.appendPaths || [];
  opts.prependPaths = opts.prependPaths || [];
  // @ts-ignore
  const paths = Module._nodeModulePaths(path.dirname(filename)); // eslint-disable-line

  // @ts-ignore
  const { parent } = Module;
  const m = new Module(filename, parent);
  m.filename = filename;
  m.paths = [].concat(opts.prependPaths).concat(paths).concat(opts.appendPaths);
  // @ts-ignore
  // eslint-disable-next-line no-underscore-dangle
  m._compile(code, filename);

  const { exports } = m;
  if (parent && parent.children) {
    parent.children.splice(parent.children.indexOf(m), 1);
  }
  return exports;
};
