import path from 'path';
import webpack from 'webpack';
import { CleanWebpackPlugin } from 'clean-webpack-plugin';
import Sass from 'sass';
import MiniCssExtractPlugin from 'mini-css-extract-plugin';
import autoprefixer from 'autoprefixer';
import { Resolver } from './resolver';
import { LoaderPresets } from '../babel/loader-preset';
import { LoaderPlugins } from '../babel/loader-plugins';
import { DEVELOPMENT, PRODUCTION } from '../constants';
import { extensionRegex, imageAssetsExtensions, staticAssetsExtensions } from './assets';
import { ConfigOptions } from '../@types/core';

const assumedRoot = process.cwd();

const serverConfig: (options: ConfigOptions) => webpack.Configuration = (
  options,
) => {
  /**
   * If the current node env is development, then set it to eval
   * for fastest rebuild and fast build, else if  the current
   * env is not production, i.e. it is something like staging
   * or test, then set it to source-map
   */
  let devtool = options.env === DEVELOPMENT ? 'eval' : false;
  if (!devtool && options.env !== PRODUCTION) {
    devtool = 'source-map';
  }
  const root = options?.root ?? assumedRoot;
  const imageAssetsRegexp = extensionRegex(imageAssetsExtensions);
  const staticAssetsRegexp = extensionRegex(staticAssetsExtensions);
  const localIdentName = options.env === PRODUCTION
    ? '[contenthash:base64:5]'
    : '[name]__[local]--[hash:base64:5]';
  let assetPrefix = options?.assetsBase;
  if (assetPrefix && !assetPrefix.endsWith('/')) {
    assetPrefix += '/';
  }
  return ({
    devtool,
    context: root,
    externals: {
      './__web_assets.json': './__web_assets.json',
    },
    externalsPresets: { node: true },
    target: 'node',
    mode: options.mode,
    resolve: Resolver.resolve,
    resolveLoader: Resolver.resolveLoader,
    stats: {
      all: false,
      colors: true,
      ids: options.verbose,
      publicPath: options.verbose,
      outputPath: true,
      assets: true,
      entrypoints: options.verbose,
      chunkGroups: options.verbose,
      chunks: options.verbose,
      modules: options.verbose,
      warnings: true,
      errors: true,
      errorStack: true,
      moduleTrace: options.verbose,
    },
    optimization: {
      // No chunks on server, it should be all combined in one file
      splitChunks: false,
    },
    entry: {
      // The entry file is the shared server index file
      server: {
        import: path.join(__dirname, '..', 'src', 'server', 'entry.ts'),
      },
    },
    // Where files should be sent once they are bundled
    output: {
      path: path.join(root, 'dist'),
      filename: 'server.js',
      publicPath: options.resourceUrl,
      library: 'server',
      // The expected should be a UMD library which can be utilized
      // in serverless applications as well.
      libraryTarget: 'umd',
      globalObject: 'this',
      umdNamedDefine: true,
      clean: true,
      assetModuleFilename: `${assetPrefix}assets/[contenthash]-[name][ext][query]`,
    },
    ignoreWarnings: [
      {
        // Ignore warnings for express
        module: /\/express\//,
      },
    ],
    // Rules of how webpack will take our files, compile & bundle them for the browser
    module: {
      rules: [
        {
          test: /\.[jt]sx?$/,
          use: {
            loader: 'babel-loader',
            options: {
              compact: false,
              presets: LoaderPresets({ node: true }),
              plugins: LoaderPlugins({ node: true }),
            },
          },
        },
        {
          test: /\.(css|s[ac]ss)$/i,
          use: [
            // Translates CSS into CommonJS
            { loader: MiniCssExtractPlugin.loader },
            {
              loader: 'css-loader',
              options: {
                modules: {
                  localIdentName,
                  mode: (resourcePath: string) => {
                    if (/pure\.(css|s[ac]ss)$/i.test(resourcePath)) {
                      return 'pure';
                    }
                    if (/global\.(css|s[ac]ss)$/i.test(resourcePath)) {
                      return 'global';
                    }
                    return 'local';
                  },
                  localIdentContext: path.resolve(root, 'src'),
                  exportLocalsConvention: 'camelCase',
                },
                importLoaders: 2,
              },
            },
            {
              loader: 'postcss-loader',
              options: {
                postcssOptions: {
                  ident: 'postcss',
                  plugins: [
                    [
                      autoprefixer,
                    ],
                  ],
                },
              },
            },
            // Compiles Sass to CSS
            {
              loader: 'sass-loader',
              options: {
                // Prefer `dart-sass`
                implementation: Sass,
                sassOptions: {
                  fiber: false,
                },
              },
            },
          ],
        },
        {
          test: imageAssetsRegexp,
          type: 'asset',
          parser: {
            dataUrlCondition: {
              maxSize: 4 * 1024, // 4kb
            },
          },
          generator: {
            // We will use content hash for long term caching of asset
            filename: `${assetPrefix}images/[contenthash]-[name][ext][query]`,
          },
        },
        {
          test: staticAssetsRegexp,
          type: 'asset/resource',
          generator: {
            // We will use content hash for long term caching of asset
            filename: `${assetPrefix}assets/[contenthash]-[name][ext][query]`,
          },
        },
        {
          resourceQuery: /raw/,
          type: 'asset/source',
        },
      ],
    },
    node: {
      __dirname: false,
      __filename: false,
    },
    plugins: [
      new webpack.EnvironmentPlugin(options?.envVars ?? {}),
      new CleanWebpackPlugin(),
      new MiniCssExtractPlugin({ filename: 'server.css' }),
    ],
  });
};

export { serverConfig };
