import path from 'path';
import fs from 'fs';
import webpack from 'webpack';
import { CleanWebpackPlugin } from 'clean-webpack-plugin';
import { BundleAnalyzerPlugin } from 'webpack-bundle-analyzer';
import MiniCssExtractPlugin from 'mini-css-extract-plugin';
import autoprefixer from 'autoprefixer';
import Sass from 'sass';
import CopyWebpackPlugin from 'copy-webpack-plugin';
import { Resolver } from './resolver';
import { LoaderPresets } from '../babel/loader-preset';
import { LoaderPlugins } from '../babel/loader-plugins';
import { DEVELOPMENT, PRODUCTION } from '../constants';
import { ConfigOptions } from '../@types/core';
import { extensionRegex, imageAssetsExtensions, staticAssetsExtensions } from './assets';
import ExtractEmittedAssets from './plugins/extract-emitted-assets';

const assumedRoot = process.cwd();

const webConfig: (options: ConfigOptions) => webpack.Configuration = (
  options,
) => {
  /**
   * If the current node env is development, then set it to eval
   * for fastest rebuild and fast build, else if  the current
   * env is not production, i.e. it is something like staging
   * or test, then set it to source-map
   */
  let devtool = options.env === DEVELOPMENT ? 'eval' : false;
  if (!devtool && options.env !== PRODUCTION) {
    devtool = 'source-map';
  }
  const root = options?.root ?? assumedRoot;
  const imageAssetsRegexp = extensionRegex(imageAssetsExtensions);
  const staticAssetsRegexp = extensionRegex(staticAssetsExtensions);
  const localIdentName = options.env === PRODUCTION
    ? '[contenthash:base64:5]'
    : '[name]__[local]--[hash:base64:5]';
  const publicPath = options?.resourceUrl ?? '/';
  const outputPath = options?.outputPath ?? path.join(root, 'dist', 'build');
  let assetPrefix = options?.assetsBase;
  let publicFolderExists = false;
  if (fs.existsSync(path.join(root, 'src', 'public'))) {
    publicFolderExists = true;
  }
  if (assetPrefix && !assetPrefix.endsWith('/')) {
    assetPrefix += '/';
  }

  const publicEnvVars: { [key: string]: any } = { ...{}, ...(options?.envVars ?? {}) };
  Object.keys(publicEnvVars).forEach(
    (key: string) => {
      if (key.startsWith('PRIVATE_')) {
        delete publicEnvVars[key];
      }
    },
  );

  return {
    devtool,
    context: root,
    target: 'web',
    mode: options.mode,
    resolve: Resolver.resolve,
    resolveLoader: Resolver.resolveLoader,
    stats: {
      all: false,
      colors: true,
      ids: options.verbose,
      publicPath: options.verbose,
      outputPath: true,
      assets: true,
      entrypoints: options.verbose,
      chunkGroups: options.verbose,
      chunks: options.verbose,
      modules: options.verbose,
      warnings: true,
      errors: true,
      errorStack: true,
      moduleTrace: options.verbose,
    },
    optimization: {
      splitChunks: {
        minSize: 20000,
        minRemainingSize: 0,
        minChunks: 1,
        maxAsyncRequests: 30,
        maxInitialRequests: 30,
        enforceSizeThreshold: 50000,
        chunks: 'async',
      },
    },
    entry: {
      client: {
        import: path.join(__dirname, '..', 'src', 'client', 'entry.ts'),
      },
    },
    // Where files should be sent once they are bundled
    output: {
      publicPath,
      path: outputPath,
      // [contenthash] for long term caching
      filename: `${assetPrefix}js/[contenthash].js`,
      assetModuleFilename: `${assetPrefix}assets/[contenthash]-[name][ext][query]`,
    },
    // Rules of how webpack will take our files, compile & bundle them for the browser
    // Rules of how webpack will take our files, compile & bundle them for the browser
    module: {
      rules: [
        {
          test: /\.[jt]sx?$/,
          use: {
            loader: 'babel-loader',
            options: {
              compact: false,
              presets: LoaderPresets(),
              plugins: LoaderPlugins(),
            },
          },
        },
        {
          test: /\.(css|s[ac]ss)$/i,
          use: [
            // Translates CSS into CommonJS
            {
              loader: options.extractCSS ? MiniCssExtractPlugin.loader : 'style-loader',
            },
            {
              loader: 'css-loader',
              options: {
                modules: {
                  localIdentName,
                  mode: (resourcePath: string) => {
                    if (/pure\.(css|s[ac]ss)$/i.test(resourcePath)) {
                      return 'pure';
                    }
                    if (/global\.(css|s[ac]ss)$/i.test(resourcePath)) {
                      return 'global';
                    }
                    return 'local';
                  },
                  localIdentContext: path.resolve(root, 'src'),
                  exportLocalsConvention: 'camelCase',
                },
                importLoaders: 2,
              },
            },
            {
              loader: 'postcss-loader',
              options: {
                postcssOptions: {
                  ident: 'postcss',
                  plugins: [
                    [
                      autoprefixer,
                    ],
                  ],
                },
              },
            },
            // Compiles Sass to CSS
            {
              loader: 'sass-loader',
              options: {
                // Prefer `dart-sass`
                implementation: Sass,
                sassOptions: {
                  fiber: false,
                },
              },
            },
          ],
        },
        {
          test: imageAssetsRegexp,
          type: 'asset',
          parser: {
            dataUrlCondition: {
              maxSize: 4 * 1024, // 4kb
            },
          },
          generator: {
            // We will use content hash for long term caching of asset
            filename: `${assetPrefix}images/[contenthash]-[name][ext][query]`,
          },
        },
        {
          test: staticAssetsRegexp,
          type: 'asset/resource',
          generator: {
            // We will use content hash for long term caching of asset
            filename: `${assetPrefix}assets/[contenthash]-[name][ext][query]`,
          },
        },
        {
          resourceQuery: /raw/,
          type: 'asset/source',
        },
      ],
    },
    ignoreWarnings: [
      (warning) => warning.toString().indexOf('__web_assets.json') !== -1,
    ],
    plugins: [
      new webpack.EnvironmentPlugin(publicEnvVars),
      new CleanWebpackPlugin(),
      ...(publicFolderExists ? [
        new CopyWebpackPlugin({
          patterns: [
            {
              from: path.join(root, 'src', 'public'),
              to: outputPath,
            },
          ],
        })] : []),
      ...(options.extractCSS ? [
        new MiniCssExtractPlugin({
          // Options similar to the same options in webpackOptions.output
          // both options are optional
          filename: `${assetPrefix}css/[contenthash].css`,
          ignoreOrder: true,
        }),
      ] : []),
      ...(options.analyze ? [new BundleAnalyzerPlugin()] : []),
      new ExtractEmittedAssets(),
    ],
  };
};

export { webConfig };
